import os
from typing import NoReturn

from elefantolib_fastapi import exceptions

from redis import asyncio as aioredis, client, utils


class Connection:
    __redis: client.Redis = None
    __aioredis: aioredis.Redis = None

    @classmethod
    def redis(cls) -> client.Redis | NoReturn:
        if redis_url := os.getenv('REDIS_URL', None):
            cls.__redis = utils.from_url(redis_url)
            return cls.__redis

        raise exceptions.RedisUrlNotFound

    @classmethod
    def aioredis(cls) -> aioredis.Redis | NoReturn:
        if redis_url := os.getenv('REDIS_URL', None):
            cls.__aioredis = aioredis.from_url(redis_url)
            return cls.__aioredis

        raise exceptions.RedisUrlNotFound

    @classmethod
    def close_redis(cls) -> None:
        if not cls.__redis:
            return None

        cls.__redis.close()

    @classmethod
    async def close_aioredis(cls) -> None:
        if not cls.__aioredis:
            return None

        await cls.__aioredis.close()



from sqlalchemy import Select, func, select
from sqlalchemy.ext.asyncio import AsyncSession


class PageNumberPagination:
    def __init__(self, session: AsyncSession, query: Select, page: int, page_size: int):
        self.session = session
        self.query = query
        self.page = page
        self.page_size = page_size

    async def get_response(self) -> dict:
        limit = self.page_size
        offset = (self.page - 1) * self.page_size

        query = self.query.limit(limit).offset(offset)
        count = await self._get_total_count()
        pages = self._get_number_of_pages(count)

        return {
            'count': count,
            'pages': pages,
            'results': [i for i in await self.session.scalars(query)],
        }

    def _get_number_of_pages(self, count: int) -> int:
        rest = count % self.page_size
        quotient = count // self.page_size

        return quotient if not rest else quotient + 1

    async def _get_total_count(self) -> int:
        return await self.session.scalar(select(func.count()).select_from(self.query.subquery()))


async def paginate(db_session, query: Select, page: int, page_size: int) -> dict:
    async with db_session as session:
        paginator = PageNumberPagination(session, query, page, page_size)
        return await paginator.get_response()



from typing import Callable

from elefantolib import context
from elefantolib.provider import fastapi_provider

import fastapi
from fastapi import routing

from .requests import Request


class APIRoute(routing.APIRoute):

    def get_route_handler(self) -> Callable:
        original_route_handler = super().get_route_handler()

        async def custom_route_handler(request: fastapi.Request) -> fastapi.Response:
            pvr = fastapi_provider.FastAPIProvider(request=request)
            request.scope['pfm'] = context.AsyncPlatformContext(pvr=pvr)
            new_request = Request(request.scope, request.receive)
            return await original_route_handler(new_request)

        return custom_route_handler



from elefantolib.context import AsyncPlatformContext

from fastapi import requests


class Request(requests.Request):

    @property
    def pfm(self) -> AsyncPlatformContext:
        return self.scope['pfm']
